@@ -9,6 +9,8 @@
  *	as published by the Free Software Foundation; either version 2.1
  *	of the License, or (at your option) any later version.
  */
+#define PAM_SM_ACCOUNT 1
+#define PAM_SM_AUTH 1
 #define PAM_SM_SESSION 1
 #define PAM_SM_PASSWORD 1
 
@@ -19,21 +21,17 @@
 #include <sys/wait.h>
 #include <assert.h>
 #include <errno.h>
+#include <pthread.h>
 #include <signal.h>
 #include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
+#include <libHX/defs.h>
+#include <libHX/proc.h>
 #include <libHX.h>
-#include "compiler.h"
-#include "misc.h"
-#include "mount.h"
 #include "pam_mount.h"
-#include "private.h"
-#include "readconfig.h"
-#include "spawn.h"
-#include "xstdlib.h"
 
 #ifndef PAM_EXTERN
 #	define PAM_EXTERN
@@ -45,9 +43,12 @@
 #	define CONFIGFILE "/etc/security/pam_mount.conf.xml"
 #endif
 
+struct pam_args {
+	bool get_pw_from_pam, get_pw_interactive, propagate_pw;
+};
+
 /* Functions */
 static void clean_config(pam_handle_t *, void *, int);
-static void clean_system_authtok(pam_handle_t *, void *, int);
 static int converse(pam_handle_t *, int, const struct pam_message **,
 	struct pam_response **);
 static int modify_pm_count(struct config *, char *, char *);
@@ -56,9 +57,8 @@ static int read_password(pam_handle_t *, const char *, char **);
 
 /* Variables */
 static const char *envpath_saved;
-unsigned int Debug = true;
-struct config Config = {};
-struct pam_args Args = {};
+struct config Config;
+struct pam_args Args;
 
 //-----------------------------------------------------------------------------
 /**
@@ -77,21 +77,27 @@ static void parse_pam_args(int argc, const char **argv)
 		assert(argv[i] != NULL);
 
 	/* first, set default values */
-	Args.auth_type  = GET_PASS;
+	Args.get_pw_from_pam    = true;
+	Args.get_pw_interactive = true;
+	Args.propagate_pw       = true;
 
 	for (i = 0; i < argc; ++i) {
-		if (strcmp("use_first_pass", argv[i]) == 0)
-			Args.auth_type = USE_FIRST_PASS;
-		else if (strcmp("try_first_pass", argv[i]) == 0)
-			Args.auth_type = TRY_FIRST_PASS;
-		else if (strcmp("soft_try_pass", argv[i]) == 0)
-			Args.auth_type = SOFT_TRY_PASS;
-		else if (strcmp("nullok", argv[i]) == 0)
-			Args.nullok = true;
-		else if (strcmp("debug", argv[i]) == 0)
+		if (strcasecmp("enable_pam_password", argv[i]) == 0)
+			Args.get_pw_from_pam = true;
+		else if (strcasecmp("disable_pam_password", argv[i]) == 0)
+			Args.get_pw_from_pam = false;
+		else if (strcasecmp("enable_interactive", argv[i]) == 0)
+			Args.get_pw_interactive = true;
+		else if (strcasecmp("disable_interactive", argv[i]) == 0)
+			Args.get_pw_interactive = false;
+		else if (strcasecmp("enable_propagate_password", argv[i]) == 0)
+			Args.propagate_pw = true;
+		else if (strcasecmp("disable_propagate_password", argv[i]) == 0)
+			Args.propagate_pw = false;
+		else if (strcasecmp("debug", argv[i]) == 0)
 			Debug = true;
 		else
-			w4rn("bad pam_mount option \"%s\"\n", argv[i]);
+			w4rn("unknown pam_mount option \"%s\"\n", argv[i]);
 	}
 }
 
@@ -156,12 +162,14 @@ static int converse(pam_handle_t *pamh, int nargs,
 	retval = pam_get_item(pamh, PAM_CONV, static_cast(const void **,
 	         static_cast(void *, &conv)));
 
-	if (retval == PAM_SUCCESS) {
+	if (retval != PAM_SUCCESS) {
+		l0g("pam_get_item: %s\n", pam_strerror(pamh, retval));
+	} else if (conv == NULL || conv->conv == NULL) {
+		w4rn("No converse function available\n");
+	} else {
 		retval = conv->conv(nargs, message, resp, conv->appdata_ptr);
 		if (retval != PAM_SUCCESS)
 			l0g("conv->conv(...): %s\n", pam_strerror(pamh, retval));
-	} else {
-		l0g("pam_get_item: %s\n", pam_strerror(pamh, retval));
 	}
 
 	if (resp == NULL || *resp == NULL || (*resp)->resp == NULL)
@@ -191,7 +199,6 @@ static int read_password(pam_handle_t *pamh, const char *prompt, char **pass)
 	assert(pamh != NULL);
 	assert(pass != NULL);
 
-	w4rn("enter read_password\n");
 	*pass = NULL;
 	msg.msg_style = PAM_PROMPT_ECHO_OFF;
 	msg.msg       = (prompt == NULL) ? "Password: " : prompt;
@@ -203,12 +210,45 @@ static int read_password(pam_handle_t *pamh, const char *prompt, char **pass)
 	return retval;
 }
 
+static void pmt_sigpipe_setup(bool block_it)
+{
+	static pthread_mutex_t sp_lock = PTHREAD_MUTEX_INITIALIZER;
+	static int sp_blocked = 0;
+	static bool sp_previous;
+	sigset_t set, oldset;
+
+	pthread_mutex_lock(&sp_lock);
+	if (block_it) {
+		if (++sp_blocked == 1) {
+			sigemptyset(&set);
+			sigaddset(&set, SIGPIPE);
+			sigprocmask(SIG_BLOCK, &set, &oldset);
+			sp_previous = sigismember(&oldset, SIGPIPE);
+		}
+	} else {
+		if (--sp_blocked == 0 && sp_previous) {
+			sigemptyset(&set);
+			sigaddset(&set, SIGPIPE);
+			sigtimedwait(&set, NULL, &(struct timespec){0, 0});
+			sigprocmask(SIG_UNBLOCK, &set, NULL);
+		}
+	}
+
+	pthread_mutex_unlock(&sp_lock);
+}
+
 static int common_init(pam_handle_t *pamh, int argc, const char **argv)
 {
 	const char *pam_user;
 	char buf[8];
 	int ret;
 
+	pmtlog_prefix = "pam_mount";
+	pmtlog_path[PMTLOG_ERR][PMTLOG_SYSLOG] = true;
+	pmtlog_path[PMTLOG_ERR][PMTLOG_STDERR] = true;
+	pmtlog_path[PMTLOG_DBG][PMTLOG_SYSLOG] = Debug;
+	pmtlog_path[PMTLOG_DBG][PMTLOG_STDERR] = Debug;
+
 	initconfig(&Config);
 	parse_pam_args(argc, argv);
 	/*
@@ -221,6 +261,8 @@ static int common_init(pam_handle_t *pamh, int argc, const char **argv)
 		/*
 		 * do NOT return %PAM_SERVICE_ERR or root will not be able to
 		 * su to other users.
+		 * Also, if we could not get the user's info, an earlier auth
+		 * module (like pam_unix2) likely blocked login already.
 		 */
 		return PAM_SUCCESS;
 	}
@@ -229,19 +271,25 @@ static int common_init(pam_handle_t *pamh, int argc, const char **argv)
 	 * disappears (valgrind)
 	 */
 	Config.user = relookup_user(pam_user);
-	if (strlen(Config.user) > MAX_PAR) {
-		l0g("username %s is too long\n", Config.user);
-		return PAM_SERVICE_ERR;
-	}
-
 	if (!readconfig(CONFIGFILE, true, &Config))
 		return PAM_SERVICE_ERR;
 
+	/* reinitialize after @Debug may have changed */
+	pmtlog_path[PMTLOG_DBG][PMTLOG_STDERR] = Debug;
+	pmtlog_path[PMTLOG_DBG][PMTLOG_SYSLOG] = Debug;
+
 	snprintf(buf, sizeof(buf), "%u", Debug);
 	setenv("_PMT_DEBUG_LEVEL", buf, true);
+
+	pmt_sigpipe_setup(true);
 	return -1;
 }
 
+static void common_exit(void)
+{
+	pmt_sigpipe_setup(false);
+}
+
 /**
  * pam_sm_authenticate -
  * @pamh:	PAM handle
@@ -251,73 +299,60 @@ static int common_init(pam_handle_t *pamh, int argc, const char **argv)
  *
  * Called by the PAM layer. The user's system password is added to PAM's
  * global module data. This is because pam_sm_open_session() does not allow
- *  access to the user's password. Returns the PAM error code or %PAM_SUCCESS.
+ * access to the user's password. Returns the PAM error code or %PAM_SUCCESS.
  */
 PAM_EXTERN EXPORT_SYMBOL int pam_sm_authenticate(pam_handle_t *pamh, int flags,
     int argc, const char **argv)
 {
 	int ret = PAM_SUCCESS;
 	char *authtok = NULL;
-	const void *tmp = NULL;
 
 	assert(pamh != NULL);
 
 	if ((ret = common_init(pamh, argc, argv)) != -1)
 		return ret;
+	w4rn(PACKAGE_STRING ": entering auth stage\n");
 
-	if (Args.auth_type != GET_PASS) { /* get password from PAM system */
+	if (Args.get_pw_from_pam) {
 		char *ptr = NULL;
+
 		ret = pam_get_item(pamh, PAM_AUTHTOK, static_cast(const void **,
 		      static_cast(void *, &ptr)));
-		if (ret != PAM_SUCCESS || ptr == NULL) {
-			if (ret == PAM_SUCCESS && ptr == NULL &&
-			    !Args.nullok)
-				ret = PAM_AUTHINFO_UNAVAIL;
-			l0g("could not get password from PAM system\n");
-			if (Args.auth_type == USE_FIRST_PASS)
-				goto out;
-		} else {
+		if (ret == PAM_SUCCESS && ptr != NULL)
 			authtok = xstrdup(ptr);
-		}
 	}
-	if (authtok == NULL) {
-		if (Args.auth_type == SOFT_TRY_PASS) {
-			ret = PAM_AUTHINFO_UNAVAIL;
-			goto out;
-		}
-		/* get password directly */
+	if (authtok == NULL && Args.get_pw_interactive) {
 		ret = read_password(pamh, Config.msg_authpw, &authtok);
-		if (ret != PAM_SUCCESS) {
-			l0g("error trying to read password\n");
-			goto out;
-		}
-		/* pam_set_item() copies to PAM-internal memory */
-		ret = pam_set_item(pamh, PAM_AUTHTOK, authtok);
-		if (ret != PAM_SUCCESS) {
-			l0g("error trying to export password\n");
-			goto out;
+		if (ret == PAM_SUCCESS && Args.propagate_pw) {
+			/*
+			 * pam_set_item() copies to PAM-internal memory.
+			 *
+			 * Using pam_set_item(PAM_AUTHTOK) here to make the
+			 * password that was just entered available to further
+			 * PAM modules.
+			 */
+			ret = pam_set_item(pamh, PAM_AUTHTOK, authtok);
+			if (ret != PAM_SUCCESS)
+				l0g("warning: failure to export password (%s)\n",
+				    pam_strerror(pamh, ret));
 		}
 	}
-	if (strlen(authtok) > MAX_PAR) {
-		l0g("password too long\n");
-		ret = PAM_AUTH_ERR;
-		goto out;
+	if (authtok != NULL) {
+		ret = pam_set_data(pamh, "pam_mount_system_authtok", authtok,
+		                   clean_system_authtok);
+		if (ret == PAM_SUCCESS) {
+			if (mlock(authtok, strlen(authtok) + 1) < 0)
+				w4rn("mlock authtok: %s\n", strerror(errno));
+		} else {
+			l0g("error trying to save authtok for session code\n");
+		}
 	}
-	w4rn("saving authtok for session code (authtok=%p)\n", authtok);
-	ret = pam_set_data(pamh, "pam_mount_system_authtok", authtok,
-	                   clean_system_authtok);
-	if (ret != PAM_SUCCESS) {
-		l0g("error trying to save authtok for session code\n");
-		goto out;
-	}
-	if (mlock(authtok, strlen(authtok) + 1) < 0)
-		w4rn("mlock authtok: %s\n", strerror(errno));
-	assert(ret != PAM_SUCCESS ||
-	       pam_get_data(pamh, "pam_mount_system_authtok", &tmp) ==
-	       PAM_SUCCESS);
-	assert(ret != PAM_SUCCESS || tmp != NULL);
- out:
-	return ret;
+	common_exit();
+	/*
+	 * pam_mount is not really meant to be an auth module. So we should not
+	 * hinder the login process.
+	 */
+	return PAM_SUCCESS;
 }
 
 /**
@@ -357,76 +392,47 @@ static void envpath_restore(void)
 static int modify_pm_count(struct config *config, char *user,
     char *operation)
 {
-	FILE *fp;
+	FILE *fp = NULL;
 	struct HXbtree *vinfo;
-	int _argc = 0, child_exit, cstdout = -1, fnval = -1;
-	const char *_argv[MAX_PAR + 1];
-	pid_t pid;
-	struct sigaction sact, oldsact;
-	unsigned int i;
+	struct HXdeque *argv;
+	struct HXproc proc;
+	int ret = -1, use_count;
 
-	assert(config_valid(config));
 	assert(user != NULL);
 	assert(operation != NULL);
 
-	/* avoid bomb on command exiting before count read */
-	memset(&sact, 0, sizeof(sact));
-	sact.sa_handler = SIG_DFL;
-	sigemptyset(&sact.sa_mask);
-	if (sigaction(SIGPIPE, &sact, &oldsact) < 0) {
-		fnval = -1;
-		goto nosigactout;
-	}
-	if ((vinfo = HXformat_init()) == NULL) {
-		fnval = -1;
+	if ((vinfo = HXformat_init()) == NULL)
 		goto out;
-	}
 	format_add(vinfo, "USER", user);
 	format_add(vinfo, "OPERATION", operation);
 	misc_add_ntdom(vinfo, user);
 
-	for (i = 0; config->command[CMD_PMVARRUN][i] != NULL; ++i)
-		add_to_argv(_argv, &_argc,
-		            config->command[CMD_PMVARRUN][i], vinfo);
-	HXformat_free(vinfo);
-	log_argv(_argv);
-
-	if (!spawn_start(_argv, &pid, NULL, &cstdout, NULL, set_myuid, NULL)) {
-		l0g("error executing pmvarrun\n");
-		fnval = -1;
+	argv = arglist_build(config->command[CMD_PMVARRUN], vinfo);
+	memset(&proc, 0, sizeof(proc));
+	proc.p_flags = HXPROC_VERBOSE | HXPROC_STDOUT;
+	proc.p_ops   = &pmt_dropprivs_ops;
+	if ((ret = pmt_spawn_dq(argv, &proc)) <= 0) {
+		l0g("error executing pmvarrun: %s\n", strerror(-ret));
 		goto out;
 	}
-	if ((fp = fdopen(cstdout, "r")) == NULL) {
-		spawn_restore_sigchld();
-		w4rn("error opening file: %s\n", strerror(errno));
-		fnval = -1;
-		goto out;
-	}
-	if (fscanf(fp, "%d", &fnval) != 1) {
-		spawn_restore_sigchld();
+	ret = -1;
+	if ((fp = fdopen(proc.p_stdout, "r")) == NULL)
+		goto out2;
+	if (fscanf(fp, "%d", &use_count) != 1)
 		w4rn("error reading login count from pmvarrun\n");
-		fnval = -1;
-		goto out;
-	}
-	if (waitpid(pid, &child_exit, 0) < 0) {
-		spawn_restore_sigchld();
-		l0g("error waiting for child: %s\n", strerror(errno));
-		fnval = -1;
-		goto out;
-	}
-	spawn_restore_sigchld();
-	if (WEXITSTATUS(child_exit)) {
-		l0g("pmvarrun failed\n");
-		fnval = -1;
-		goto out;
-	}
-	w4rn("pmvarrun says login count is %d\n", fnval);
+	else
+		w4rn("pmvarrun says login count is %d\n", use_count);
+ out2:
+	if (fp != NULL)
+		fclose(fp);
+	else
+		close(proc.p_stdout);
+	if (HXproc_wait(&proc) >= 0 && proc.p_exited && proc.p_status == 0)
+		ret = use_count;
  out:
-	sigaction(SIGPIPE, &oldsact, NULL);
- nosigactout:
-
-	assert(config_valid(config));
-	return fnval;
+	if (vinfo != NULL)
+		HXformat_free(vinfo);
+	return ret;
 }
 
 /**
@@ -446,7 +452,7 @@ PAM_EXTERN EXPORT_SYMBOL int pam_sm_open_session(pam_handle_t *pamh, int flags,
 	struct vol *vol;
 	int ret;
 	unsigned int krb5_set;
-	char *system_authtok;
+	char *system_authtok = NULL;
 	const void *tmp;
 	int getval;
 
@@ -455,7 +461,7 @@ PAM_EXTERN EXPORT_SYMBOL int pam_sm_open_session(pam_handle_t *pamh, int flags,
 	if ((ret = common_init(pamh, argc, argv)) != -1)
 		return ret;
 
-	w4rn("Entered pam_mount session stage\n");
+	w4rn(PACKAGE_STRING ": entering session stage\n");
 
 	/*
 	 * Get the Kerberos CCNAME so we can make it available to the
@@ -476,16 +482,16 @@ PAM_EXTERN EXPORT_SYMBOL int pam_sm_open_session(pam_handle_t *pamh, int flags,
 		}
 	}
 
-	w4rn("back from global readconfig\n");
-	if (strlen(Config.luserconf) == 0)
-		w4rn("per-user configurations not allowed by pam_mount.conf.xml\n");
-	else if (exists(Config.luserconf) && owns(Config.user, Config.luserconf)) {
-		w4rn("going to readconfig user\n");
+	if (Config.luserconf == NULL || strlen(Config.luserconf) == 0)
+		;
+	else if (!pmt_fileop_exists(Config.luserconf))
+		;
+	else if (pmt_fileop_owns(Config.user, Config.luserconf)) {
+		w4rn("going to readconfig %s\n", Config.luserconf);
 		if (!readconfig(Config.luserconf, false, &Config)) {
 			ret = PAM_SERVICE_ERR;
 			goto out;
 		}
-		w4rn("back from user readconfig\n");
 	} else
 		w4rn("%s does not exist or is not owned by user\n",
 		     Config.luserconf);
@@ -503,17 +509,19 @@ PAM_EXTERN EXPORT_SYMBOL int pam_sm_open_session(pam_handle_t *pamh, int flags,
 	ret = pam_get_data(pamh, "pam_mount_system_authtok",
 	      static_cast(const void **, static_cast(void *, &system_authtok)));
 	if (ret != PAM_SUCCESS) {
-		if (Args.auth_type == SOFT_TRY_PASS) {
-			ret = PAM_AUTHINFO_UNAVAIL;
-			goto out;
-		}
-		l0g("error trying to retrieve authtok from auth code\n");
-		ret = read_password(pamh, Config.msg_sessionpw, &system_authtok);
-		if (ret != PAM_SUCCESS) {
-			l0g("error trying to read password\n");
-			goto out;
+		if (Args.get_pw_interactive) {
+			ret = read_password(pamh, Config.msg_sessionpw, &system_authtok);
+			if (ret != PAM_SUCCESS)
+				l0g("warning: could not obtain password "
+				    "interactively either\n");
 		}
+		/*
+		 * Proceed without a password. Some volumes may not need one,
+		 * e.g. bind mounts and networked/unencrypted volumes.
+		 */
 	}
+	if (system_authtok == NULL)
+		system_authtok = xstrdup("");
 
 	misc_dump_id("Session open");
 
@@ -530,20 +538,34 @@ PAM_EXTERN EXPORT_SYMBOL int pam_sm_open_session(pam_handle_t *pamh, int flags,
 		if (!vol->globalconf &&
 		    !luserconf_volume_record_sane(&Config, vol))
 			continue;
-		w4rn("about to perform mount operations\n");
 
 		if (!mount_op(do_mount, &Config, vol, system_authtok)) {
-			l0g("mount of %s failed\n", vol->volume);
+			l0g("mount of %s failed\n", znul(vol->volume));
 			ret = PAM_SERVICE_ERR;
 		}
 	}
-	clean_system_authtok(pamh, system_authtok, 0);
+	memset(system_authtok, 0, strlen(system_authtok));
 	if (krb5_set)
 		unsetenv("KRB5CCNAME");
 	modify_pm_count(&Config, Config.user, "1");
 	envpath_restore();
+	if (getuid() == 0)
+		/* Make sure root can always log in. */
+		/* NB: I don't even wanna think of SELINUX's ambiguous UIDs... */
+		ret = PAM_SUCCESS;
+
+	/*
+	 * If mounting something failed, e.g. ret = %PAM_SERVICE_ERR, we have
+	 * to unravel everything and umount all volumes. But *only* if
+	 * pam_mount was configured as a "required" module. How can this info
+	 * be obtained?
+	 * For now, always assume "optional", so that the volumes are
+	 * definitely unmounted when the user logs out again.
+	 */
+	ret = PAM_SUCCESS;
  out:
 	w4rn("done opening session (ret=%d)\n", ret);
+	common_exit();
 	return ret;
 }
 
@@ -580,7 +602,6 @@ PAM_EXTERN EXPORT_SYMBOL int pam_sm_close_session(pam_handle_t *pamh,
     int flags, int argc, const char **argv)
 {
 	int ret = PAM_SUCCESS;
-	struct vol *vol;
 	const char *pam_user = NULL;
 
 	assert(pamh != NULL);
@@ -599,10 +620,6 @@ PAM_EXTERN EXPORT_SYMBOL int pam_sm_close_session(pam_handle_t *pamh,
 	ret = pam_get_user(pamh, &pam_user, NULL);
 	if (ret != PAM_SUCCESS) {
 		l0g("could not get user\n");
-		/*
-		 * do NOT return %PAM_SERVICE_ERR or root will not be able to
-		 * su to other users.
-		 */
 		goto out;
 	}
 	/*
@@ -615,17 +632,12 @@ PAM_EXTERN EXPORT_SYMBOL int pam_sm_close_session(pam_handle_t *pamh,
 		l0g("could not chdir\n");
 
 	envpath_init(Config.path);
-	if (modify_pm_count(&Config, Config.user, "-1") <= 0) {
-		HXlist_for_each_entry_rev(vol, &Config.volume_list, list) {
-			w4rn("going to unmount\n");
-			if (!mount_op(do_unmount, &Config, vol, NULL))
-				l0g("unmount of %s failed\n",
-				    vol->volume);
-		}
-	} else {
+	if (modify_pm_count(&Config, Config.user, "-1") > 0)
 		w4rn("%s seems to have other remaining open sessions\n",
 		     Config.user);
-	}
+	else
+		umount_final(&Config);
+
 	envpath_restore();
  out:
 	/*
