@@ -6,7 +6,6 @@
  *	as published by the Free Software Foundation; either version 2.1
  *	of the License, or (at your option) any later version.
  */
-#include <ctype.h>
 #include <errno.h>
 #include <grp.h>
 #include <pwd.h>
@@ -21,14 +20,13 @@
 #	include <mntent.h>
 #endif
 #include <libHX/arbtree.h>
+#include <libHX/ctype_helper.h>
+#include <libHX/defs.h>
+#include <libHX/deque.h>
 #include <libHX/option.h>
 #include <libHX/string.h>
-#include "compiler.h"
-#include "misc.h"
+#include <libHX/libxml_helper.h>
 #include "pam_mount.h"
-#include "private.h"
-#include "readconfig.h"
-#include "xstdlib.h"
 
 /* Definitions */
 enum {
@@ -47,6 +45,10 @@ enum {
 	CMDA_AUTHPW,
 	CMDA_SESSIONPW,
 	CMDA_PATH,
+	CMDA_LOGOUT_SIGHUP,
+	CMDA_LOGOUT_SIGTERM,
+	CMDA_LOGOUT_SIGKILL,
+	CMDA_LOGOUT_WAIT,
 };
 
 enum {
@@ -62,48 +64,155 @@ struct callbackmap {
 
 struct pmt_command {
 	enum command_type type;
-	const char *fs, *command_name, *def[MAX_PAR + 1];
+	const char *fs, *def[9];
+	/*
+	 * You will need to enlarge @def whenever the compiler
+	 * (rightfully) complains.
+	 */
 };
 
 /* Functions */
-static char *expand_home(const char *, char *, size_t);
-static char *expand_user(const char *, char *, size_t);
-static inline int strcmp_1u(const xmlChar *, const char *);
 static int rc_volume_cond_ext(const struct passwd *, xmlNode *);
 
 /* Variables */
 static const struct callbackmap cf_tags[];
 static const struct pmt_command default_command[];
-static bool onetime_options_allow, onetime_options_require;
 
 //-----------------------------------------------------------------------------
+/**
+ * expand_home -
+ * @user:	username to use for home directory lookup
+ * @path:	pathname to expand
+ * @size:	size of @path
+ *
+ * Expands tildes in @path to the user home directory and updates @path.
+ * Returns @dest.
+ */
+static bool expand_home(const char *user, char **path_pptr)
+{
+	char *buf, *path = *path_pptr;
+	struct passwd *pe;
+	size_t size;
+
+	if (path == NULL)
+		return true;
+	if (*path != '~')
+		return true;
+	if ((pe = getpwnam(user)) == NULL) {
+		l0g("Could not lookup account info for %s\n", user);
+		return false;
+	}
+	size = strlen(pe->pw_dir) + strlen(path) + 1;
+	if ((buf = xmalloc(size)) == NULL) {
+		l0g("%s\n", strerror(errno));
+		return NULL;
+	}
+	snprintf(buf, size, "%s%s", pe->pw_dir, path + 1);
+	free(path);
+	*path_pptr = buf;
+	return true;
+}
+
+/**
+ * expand_user -
+ * @user:	username to substitue for placeholder
+ * @dest:	buffer to operate on
+ * @size:	size of @dest
+ *
+ * Substitutes all occurrences of %(USER) by the username. Returns NULL on
+ * failure, otherwise @dest.
+ *
+ * (This should probably be done by the fmt_ptrn stuff, but is not at the
+ * moment due to to-XML transition period.)
+ */
+static bool expand_user(const char *user, char **dest_pptr,
+    const struct HXbtree *vinfo)
+{
+	hxmc_t *tmp = NULL;
+
+	if (*dest_pptr == NULL)
+		return true;
+	HXformat_aprintf(vinfo, &tmp, *dest_pptr);
+	*dest_pptr = xstrdup(tmp);
+	HXmc_free(tmp);
+	return true;
+}
+
 /**
  * expandconfig -
  * @config:	configuration structure
  *
- * Expands all wildcards in the structure.
+ * This function will expand variables in attributes itself.
+ * Like, turning <volume server="w2k3" path="%(USER)"
+ * mountpoint="~" /> into path="joe" mountpoint="/home/joe".
  */
 bool expandconfig(const struct config *config)
 {
 	const char *u = config->user;
+	struct HXbtree *vinfo;
+	struct passwd *pe;
+	struct group *ge;
+	struct kvp *kvp;
 	struct vol *vpt;
 
-	HXlist_for_each_entry(vpt, &config->volume_list, list) {
-		if (expand_home(u, vpt->mountpoint, sizeof(vpt->mountpoint)) == NULL ||
-		    expand_user(u, vpt->mountpoint, sizeof(vpt->mountpoint)) == NULL ||
-		    expand_home(u, vpt->volume, sizeof(vpt->volume)) == NULL ||
-		    expand_user(u, vpt->volume, sizeof(vpt->volume)) == NULL ||
-		    expand_home(u, vpt->fs_key_path, sizeof(vpt->fs_key_path)) == NULL ||
-		    expand_user(u, vpt->fs_key_path, sizeof(vpt->fs_key_path)) == NULL)
-			return false;
-
-		if (strcmp(vpt->user, "*") == 0 || *vpt->user == '@')
-			vpt->used_wildcard = true;
-
-		strcpy(vpt->user, config->user);
+	if ((vinfo = HXformat_init()) == NULL)
+		return false;
+	if ((pe = getpwnam(u)) == NULL) {
+		l0g("You do not exist? %s? %s.\n", u, strerror(errno));
+		goto rfalse;
 	}
 
+	HXformat_add(vinfo, "USER", u, HXTYPE_STRING);
+	HXformat_add(vinfo, "USERUID", reinterpret_cast(void *,
+		static_cast(long, pe->pw_uid)), HXTYPE_UINT | HXFORMAT_IMMED);
+	HXformat_add(vinfo, "USERGID", reinterpret_cast(void *,
+		static_cast(long, pe->pw_gid)), HXTYPE_UINT | HXFORMAT_IMMED);
+	ge = getgrgid(pe->pw_gid);
+	format_add(vinfo, "GROUP", (ge != NULL) ? ge->gr_name : "");
+	misc_add_ntdom(vinfo, u);
+
+	HXlist_for_each_entry(vpt, &config->volume_list, list) {
+		if (!expand_user(u, &vpt->server, vinfo) ||
+		    !expand_home(u, &vpt->volume) ||
+		    !expand_user(u, &vpt->volume, vinfo) ||
+		    !expand_home(u, &vpt->mountpoint) ||
+		    !expand_user(u, &vpt->mountpoint, vinfo) ||
+		    !expand_home(u, &vpt->fs_key_path) ||
+		    !expand_user(u, &vpt->fs_key_path, vinfo) ||
+		    !expand_user(u, &vpt->fs_key_cipher, vinfo))
+			goto rfalse;
+
+		/*
+		 * Substitution on the key seems odd indeed, and was not
+		 * originally intended. But it should be done, since the
+		 * option separator is notalways ",", e.g. with AUFS.
+		 * Least Surprise says: do the expansion on the key
+		 * nevertheless.
+		 */
+		HXlist_for_each_entry(kvp, &vpt->options, list)
+			if (!expand_user(u, &kvp->key, vinfo) ||
+			    !expand_user(u, &kvp->value, vinfo))
+				goto rfalse;
+	}
+
+	HXformat_free(vinfo);
 	return true;
+ rfalse:
+	HXformat_free(vinfo);
+	return false;
+}
+
+static void volume_free(struct vol *vol)
+{
+	kvplist_genocide(&vol->options);
+	free(vol->fstype);
+	free(vol->server);
+	free(vol->volume);
+	free(vol->cipher);
+	free(vol->fs_key_cipher);
+	free(vol->fs_key_hash);
+	free(vol->fs_key_path);
+	free(vol);
 }
 
 /**
@@ -114,19 +223,31 @@ bool expandconfig(const struct config *config)
  */
 void freeconfig(struct config *config)
 {
+	struct HXdeque *cmd;
 	struct vol *vol, *next;
-	unsigned int i, j;
+	unsigned int i;
+
+	HXmc_free(config->luserconf);
 
 	for (i = 0; i < _CMD_MAX; ++i) {
-		free(config->command[i][0]);
-		for (j = 0; config->command[i][j] != NULL; ++j)
-			config->command[i][j] = NULL;
+		/*
+		 * This comment shall denote the awareness of a slight memory
+		 * leak. In initconfig(), all strings of a command are
+		 * duplicated when added to the list, while here we only free
+		 * the first element. Because if the default hardcoded argument
+		 * list is overriden using .conf.xml, only the first element is
+		 * allocated and the rest are pointers to the same memory
+		 * block.
+		 */
+		if ((cmd = config->command[i]) == NULL)
+			continue;
+		if (cmd->items > 0)
+			free(cmd->first->ptr);
+		HXdeque_free(cmd);
 	}
 
-	HXlist_for_each_entry_safe(vol, next, &config->volume_list, list) {
-		kvplist_genocide(&vol->options);
-		free(vol);
-	}
+	HXlist_for_each_entry_safe(vol, next, &config->volume_list, list)
+		volume_free(vol);
 
 	HXbtree_free(config->options_allow);
 	HXbtree_free(config->options_require);
@@ -135,6 +256,7 @@ void freeconfig(struct config *config)
 	free(config->msg_authpw);
 	free(config->msg_sessionpw);
 	free(config->path);
+	memset(config, 0, sizeof(*config));
 }
 
 /**
@@ -208,23 +330,35 @@ void initconfig(struct config *config)
 {
 	unsigned int i, j;
 	char options_allow[]   = "nosuid,nodev";
-	char options_require[] = "nosuid,nodev";;
+	char options_require[] = "nosuid,nodev";
 
 	memset(config, 0, sizeof(*config));
 	config->debug      = true;
 	config->mkmntpoint = true;
-	strcpy(config->fsckloop, "/dev/loop7");
 
 	config->msg_authpw    = xstrdup("pam_mount password:");
 	config->msg_sessionpw = xstrdup("reenter password for pam_mount:");
 
-	config->path = xstrdup("/sbin:/bin:/usr/sbin:/usr/bin:"
-	               "/usr/local/sbin:/usr/local/bin");
+	config->path = xstrdup(PMT_DFL_PATH);
+
+	/* Initialize all. Makes it easier. */
+	for (i = 0; i < _CMD_MAX; ++i)
+		if ((config->command[i] = HXdeque_init()) == NULL)
+			perror("malloc");
+
+	for (i = 0; default_command[i].type != -1; ++i) {
+		struct HXdeque *cmd = config->command[default_command[i].type];
+
+		if (cmd->items > 0)
+			/*
+			 * Already initialized by an earlier entry
+			 * with same type.
+			 */
+			continue;
 
-	for (i = 0; default_command[i].type != -1; ++i)
 		for (j = 0; default_command[i].def[j] != NULL; ++j)
-			config->command[default_command[i].type][j] =
-				xstrdup(default_command[i].def[j]);
+			HXdeque_push(cmd, xstrdup(default_command[i].def[j]));
+	}
 
 	config->options_allow   = HXbtree_init(OPT_TREE_FLAGS);
 	config->options_require = HXbtree_init(OPT_TREE_FLAGS);
@@ -241,19 +375,22 @@ bool readconfig(const char *file, bool global_conf, struct config *config)
 	xmlDoc *doc;
 	xmlNode *ptr;
 
-	if ((doc = xmlParseFile(file)) == NULL)
+	if ((doc = xmlParseFile(file)) == NULL) {
+		l0g("libxml detected a syntax error in %s\n", file);
 		return false;
+	}
 	ptr = xmlDocGetRootElement(doc);
-	if (ptr == NULL || strcmp_1u(ptr->name, "pam_mount") != 0) {
+	if (ptr == NULL || xml_strcmp(ptr->name, "pam_mount") != 0) {
 		xmlFreeDoc(doc);
 		return false;
 	}
 
+	config->level = global_conf ? CONTEXT_GLOBAL : CONTEXT_LUSER;
 	for (ptr = ptr->children; ptr != NULL; ptr = ptr->next) {
 		if (ptr->type != XML_ELEMENT_NODE)
 			continue;
 		for (cmp = cf_tags; cmp->name != NULL; ++cmp)
-			if (strcmp_1u(ptr->name, cmp->name) == 0) {
+			if (xml_strcmp(ptr->name, cmp->name) == 0) {
 				err = (*cmp->func)(ptr, config, cmp->cmd);
 				if (err != NULL)
 					l0g("%s\n", err);
@@ -266,89 +403,24 @@ bool readconfig(const char *file, bool global_conf, struct config *config)
 }
 
 //-----------------------------------------------------------------------------
-/**
- * expand_home -
- * @user:	username to use for home directory lookup
- * @path:	pathname to expand
- * @size:	size of @path
- *
- * Expands tildes in @path to the user home directory and updates @path.
- * Returns @dest.
- */
-static char *expand_home(const char *user, char *path, size_t size)
-{
-	struct passwd *pe;
-	char *buf;
-
-	if (path[0] != '~' || path[1] != '/')
-		return path;
-
-	if ((pe = getpwnam(user)) == NULL) {
-		l0g("Could not lookup account information for %s\n", user);
-		return NULL;
-	}
-
-	if ((buf = xmalloc(size)) == NULL) {
-		l0g("%s\n", strerror(errno));
-		return NULL;
-	}
-	if (snprintf(buf, size, "%s%s", pe->pw_dir, path + 1) >= size)
-		l0g("Warning: Not enough buffer space in expand_home()\n");
-
-	strncpy(path, buf, size);
-	free(buf);
-	return path;
-}
-
-/**
- * expand_user -
- * @user:	username to substitue for placeholder
- * @dest:	buffer to operate on
- * @size:	size of @dest
- *
- * Substitutes all occurrences of %(USER) by the username. Returns NULL on
- * failure, otherwise @dest.
- *
- * (This should probably be done by the fmt_ptrn stuff, but is not at the
- * moment due to to-XML transition period.)
- */
-static char *expand_user(const char *user, char *dest, size_t size)
-{
-	struct HXbtree *vinfo;
-
-	if (dest == NULL)
-		l0g("expand_user_wildcard(dest=NULL), please fix\n");
-
-	if ((vinfo = HXformat_init()) == NULL)
-		return NULL;
-	HXformat_add(vinfo, "USER", user, HXTYPE_STRING);
-	misc_add_ntdom(vinfo, user);
-	HXformat_sprintf(vinfo, dest, size, dest);
-	HXformat_free(vinfo);
-	return dest;
-}
-
 /**
  * fstab_value -
  * @volume:	path to volume
  * @field:	-
- * @dest:	destination buffer
- * @size:	size of @dest
  *
  * Search for @volume in /etc/fstab and if it is found, copy the @field'th
- * field to @dest which is of size @size. Returns 0 on error, 1 on success.
+ * field and return it. Returns %NULL on error.
  */
-static int fstab_value(const char *volume, const enum fstab_field field,
-    char *dest, int size)
+static char *fstab_value(const char *volume, const enum fstab_field field)
 {
-	const char *val;
+	char *val;
 #if defined(__linux__)
 	struct mntent *fstab_record;
 	FILE *fstab;
 
 	if ((fstab = setmntent("/etc/fstab", "r")) == NULL) {
 		l0g("could not open fstab\n");
-		return 0;
+		return NULL;
 	}
 
 	for (fstab_record = getmntent(fstab);
@@ -359,7 +431,7 @@ static int fstab_value(const char *volume, const enum fstab_field field,
 
 	if (fstab_record == NULL) {
 		l0g("could not get %dth fstab field for %s\n", field, volume);
-		return 0;
+		return NULL;
 	}
 
 	switch (field) {
@@ -377,18 +449,18 @@ static int fstab_value(const char *volume, const enum fstab_field field,
 			break;
 		default:
 			l0g("field of %d invalid\n", field);
-			return 0;
+			return NULL;
 	}
 #elif defined (__FreeBSD__) || defined (__OpenBSD__) || defined(__APPLE__)
 	struct fstab *fstab_record;
 
 	if (!setfsent()) {
 		l0g("could not open fstab\n");
-		return 0;
+		return NULL;
 	}
 	if ((fstab_record = getfsspec(volume)) == NULL) {
 		l0g("could not get %dth fstab field for %s\n", field, volume);
-		return 0;
+		return NULL;
 	}
 
 	switch (field) {
@@ -406,21 +478,20 @@ static int fstab_value(const char *volume, const enum fstab_field field,
 			break;
 		default:
 			l0g("field of %d invalid\n", field);
-			return 0;
+			return NULL;
 	}
 #else
 	l0g("reading fstab not implemented on arch.\n");
-	return 0;
+	return NULL;
 #endif
 
-	strncpy(dest, val, size);
-	dest[size-1] = '\0';
+	val = xstrdup(val);
 #if defined(__linux__)
 	endmntent(fstab);
 #elif defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__APPLE__)
 	endfsent();
 #endif
-	return 1;
+	return val;
 }
 
 /**
@@ -437,12 +508,12 @@ static inline char *get_next_argument(char **sptr)
 
 	if (*i == '\0')
 		return NULL;
-	while (isspace(*i))
+	while (HX_isspace(*i))
 		++i;
 
 	while (*i != '\0') {
 		if (quot == '\0') {
-			if (isspace(*i)) {
+			if (HX_isspace(*i)) {
 				++i;
 				break;
 			}
@@ -493,11 +564,6 @@ static inline bool parse_bool_f(char *s)
 	return ret;
 }
 
-static inline int strcmp_1u(const xmlChar *a, const char *b)
-{
-	return strcmp(reinterpret_cast(const char *, a), b);
-}
-
 /**
  * user_in_sgrp -
  * @user:	user to check
@@ -510,17 +576,17 @@ static inline int strcmp_1u(const xmlChar *a, const char *b)
 static bool user_in_sgrp(const char *user, const char *grp, bool icase)
 {
 	struct group *gent;
-	const char **wp;
+	const char *const *wp;
 
 	if ((gent = getgrnam(grp)) == NULL) {
 		w4rn("getgrnam(\"%s\") failed: %s\n", grp, strerror(errno));
 		return false;
 	}
 
-	wp = const_cast(const char **, gent->gr_mem);
+	wp = const_cast2(const char *const *, gent->gr_mem);
 	while (wp != NULL && *wp != NULL) {
-		if (strcmp(*wp, user) ||
-		    (icase && strcasecmp(*wp, user)) == 0)
+		if (strcmp(*wp, user) == 0 ||
+		    (icase && strcasecmp(*wp, user) == 0))
 			return true;
 		++wp;
 	}
@@ -528,23 +594,33 @@ static bool user_in_sgrp(const char *user, const char *grp, bool icase)
 	return false;
 }
 
-static inline char *xmlGetProp_2s(xmlNode *node, const char *attr)
-{
-	return reinterpret_cast(char *, xmlGetProp(node,
-	       reinterpret_cast(const xmlChar *, attr)));
-}
-
 //-----------------------------------------------------------------------------
 static const char *rc_command(xmlNode *node, struct config *config,
-    unsigned int command)
+    unsigned int cmdnr)
 {
-	unsigned int n = 0;
+	struct HXdeque *this_cmd;
 	char *arg, *wp;
 
 	if (config->level != CONTEXT_GLOBAL)
 		return "Tried to set command from user config\n";
 	if ((node = node->children) == NULL)
 		return NULL;
+
+	/*
+	 * If the command was previously defined, delete the first node. The
+	 * stored pointer is returned and we free that. Since the other
+	 * pointers are not an allocation head, we only free their nodes.
+	 * List has zero elements afterwards.
+	 */
+	if ((this_cmd = config->command[cmdnr]) != NULL &&
+	    this_cmd->items > 0) {
+		free(HXdeque_del(this_cmd->first));
+		HXdeque_free(this_cmd);
+		this_cmd = NULL;
+	}
+	if (this_cmd == NULL)
+		this_cmd = config->command[cmdnr] = HXdeque_init();
+
 	for (; node != NULL; node = node->next) {
 		if (node->type != XML_TEXT_NODE)
 			continue;
@@ -557,7 +633,7 @@ static const char *rc_command(xmlNode *node, struct config *config,
 			 * @config->command[command][0] will be the pointer to
 			 * the block to free later.
 			 */
-			config->command[command][n++] = arg;
+			HXdeque_push(this_cmd, arg);
 
 		/* No hassle to support comment-split tags. */
 		break;
@@ -569,28 +645,28 @@ static const char *rc_debug(xmlNode *node, struct config *config,
     unsigned int cmd)
 {
 	char *s;
-	if ((s = xmlGetProp_2s(node, "enable")) != NULL)
+	if ((s = xml_getprop(node, "enable")) != NULL)
 		Debug = config->debug = strtoul(s, NULL, 0);
+
+		pmtlog_path[PMTLOG_DBG][PMTLOG_SYSLOG] =
+		pmtlog_path[PMTLOG_DBG][PMTLOG_STDERR] =
+			Debug;
 	free(s);
 	return NULL;
 }
 
-static const char *rc_fsckloop(xmlNode *node, struct config *config,
-    unsigned int cmd)
+static const char *rc_logout(xmlNode *node, struct config *config,
+    unsigned int command)
 {
-	char *dev;
+	char *tmp;
 
-	if (config->level != CONTEXT_GLOBAL)
-		return "Tried to set <fsckloop> from user config";
-	if ((dev = xmlGetProp_2s(node, "device")) != NULL) {
-		if (strlen(dev) > PATH_MAX) {
-			free(dev);
-			return "fsckloop device path too long";
-		}
-		strncpy(config->fsckloop, dev, PATH_MAX);
-		free(dev);
+	if ((tmp = xml_getprop(node, "wait")) != NULL) {
+		config->sig_wait = strtoul(tmp, NULL, 0);
+		free(tmp);
 	}
-
+	config->sig_hup  = parse_bool_f(xml_getprop(node, "hup"));
+	config->sig_term = parse_bool_f(xml_getprop(node, "term"));
+	config->sig_kill = parse_bool_f(xml_getprop(node, "kill"));
 	return NULL;
 }
 
@@ -604,15 +680,12 @@ static const char *rc_luserconf(xmlNode *node, struct config *config,
 		return "Tried to set <luserconf> from user config";
 	if ((pent = getpwnam(config->user)) == NULL)
 		return "Could not get password entry";
-	if ((s = xmlGetProp_2s(node, "name")) == NULL)
+	if ((s = xml_getprop(node, "name")) == NULL)
 		return "<luserconf> is missing name= attribute";
-	if (strlen(pent->pw_dir) + 1 + strlen(s) > sizeof_z(config->luserconf)) {
-		free(s);
-		return "expanded luserconf path too long";
-	}
-	HX_strlcpy(config->luserconf, pent->pw_dir, sizeof(config->luserconf));
-	HX_strlcat(config->luserconf, "/", sizeof(config->luserconf));
-	HX_strlcat(config->luserconf, s, sizeof(config->luserconf));
+	HXmc_free(config->luserconf);
+	config->luserconf = HXmc_strinit(pent->pw_dir);
+	HXmc_strcat(&config->luserconf, "/");
+	HXmc_strcat(&config->luserconf, s);
 	w4rn("path to luserconf set to %s\n", config->luserconf);
 	free(s);
 	return NULL;
@@ -622,10 +695,10 @@ static const char *rc_mkmountpoint(xmlNode *node, struct config *config,
     unsigned int command)
 {
 	char *s;
-	if ((s = xmlGetProp_2s(node, "enable")) != NULL)
+	if ((s = xml_getprop(node, "enable")) != NULL)
 		config->mkmntpoint = strtol(s, NULL, 0);
 	free(s);
-	if ((s = xmlGetProp_2s(node, "remove")) != NULL)
+	if ((s = xml_getprop(node, "remove")) != NULL)
 		config->rmdir_mntpt = parse_bool(s);
 	free(s);
 	return NULL;
@@ -640,11 +713,11 @@ static const char *rc_mntoptions(xmlNode *node, struct config *config,
 	if (config->level != CONTEXT_GLOBAL)
 		return "Tried to set <mntoptions allow=...> from user config";
 
-	if ((options = xmlGetProp_2s(node, "allow")) != NULL) {
-		if (!onetime_options_allow) {
+	if ((options = xml_getprop(node, "allow")) != NULL) {
+		if (!config->seen_mntoptions_allow) {
 			HXbtree_free(config->options_allow);
 			config->options_allow = HXbtree_init(OPT_TREE_FLAGS);
-			onetime_options_allow = true;
+			config->seen_mntoptions_allow = true;
 		}
 		ret = str_to_optlist(config->options_allow, options);
 		free(options);
@@ -652,18 +725,23 @@ static const char *rc_mntoptions(xmlNode *node, struct config *config,
 			return "Error parsing allowed options";
 	}
 
-	if ((options = xmlGetProp_2s(node, "deny")) != NULL) {
+	if ((options = xml_getprop(node, "deny")) != NULL) {
 		ret = str_to_optlist(config->options_deny, options);
 		free(options);
 		if (!ret)
 			return "Error parsing denied options";
 	}
 
-	if ((options = xmlGetProp_2s(node, "require")) != NULL) {
-		if (!onetime_options_require) {
+	if ((options = xml_getprop(node, "require")) != NULL) {
+		/*
+		 * On the first sight of "require" in the config file, the
+		 * built-in default shall be overridden, but multiple
+		 * "require"s shall be cumulative.
+		 */
+		if (!config->seen_mntoptions_require) {
 			HXbtree_free(config->options_require);
 			config->options_require = HXbtree_init(OPT_TREE_FLAGS);
-			onetime_options_require = true;
+			config->seen_mntoptions_require = true;
 		}
 		ret = str_to_optlist(config->options_require, options);
 		free(options);
@@ -841,11 +919,11 @@ static int rc_volume_cond_user(const struct passwd *pwd, xmlNode *node)
 	for (node = node->children; node != NULL; node = node->next) {
 		if (node->type != XML_TEXT_NODE)
 			continue;
-		if (parse_bool_f(xmlGetProp_2s(parent, "icase")))
+		if (parse_bool_f(xml_getprop(parent, "icase")))
 			return strcasecmp(signed_cast(const char *,
 			       node->content), pwd->pw_name) == 0;
 		else
-			return strcmp_1u(node->content, pwd->pw_name) == 0;
+			return xml_strcmp(node->content, pwd->pw_name) == 0;
 	}
 
 	return false;
@@ -947,7 +1025,7 @@ static int rc_volume_cond_pgrp(const struct passwd *pwd, xmlNode *node)
 
 		return __rc_volume_cond_pgrp(
 		       signed_cast(const char *, node->content), pwd->pw_gid,
-		       parse_bool_f(xmlGetProp_2s(parent, "icase")));
+		       parse_bool_f(xml_getprop(parent, "icase")));
 	}
 
 	l0g("config: empty or invalid content for <%s>\n", "pgrp");
@@ -978,9 +1056,10 @@ static int rc_volume_cond_sgrp(const struct passwd *pwd, xmlNode *node)
 		ret = rc_volume_cond_pgrp(pwd, parent);
 		if (ret < 0 || ret > 0)
 			return ret;
+		const char *nnn = signed_cast(const char *, node->content);
 		return user_in_sgrp(pwd->pw_name,
-		       signed_cast(const char *, node->content),
-		       parse_bool_f(xmlGetProp_2s(parent, "icase")));
+		       (void*)nnn,
+		       parse_bool_f(xml_getprop(parent, "icase")));
 	}
 
 	l0g("config: empty or invalid content for <%s>\n", "sgrp");
@@ -994,23 +1073,23 @@ static int rc_volume_cond_sgrp(const struct passwd *pwd, xmlNode *node)
  */
 static int rc_volume_cond_ext(const struct passwd *pwd, xmlNode *node)
 {
-	if (strcmp_1u(node->name, "and") == 0)
+	if (xml_strcmp(node->name, "and") == 0)
 		return rc_volume_cond_and(pwd, node);
-	else if (strcmp_1u(node->name, "or") == 0)
+	else if (xml_strcmp(node->name, "or") == 0)
 		return rc_volume_cond_or(pwd, node);
-	else if (strcmp_1u(node->name, "xor") == 0)
+	else if (xml_strcmp(node->name, "xor") == 0)
 		return rc_volume_cond_xor(pwd, node);
-	else if (strcmp_1u(node->name, "not") == 0)
+	else if (xml_strcmp(node->name, "not") == 0)
 		return rc_volume_cond_not(pwd, node);
-	else if (strcmp_1u(node->name, "user") == 0)
+	else if (xml_strcmp(node->name, "user") == 0)
 		return rc_volume_cond_user(pwd, node);
-	else if (strcmp_1u(node->name, "uid") == 0)
+	else if (xml_strcmp(node->name, "uid") == 0)
 		return rc_volume_cond_uid(pwd, node);
-	else if (strcmp_1u(node->name, "gid") == 0)
+	else if (xml_strcmp(node->name, "gid") == 0)
 		return rc_volume_cond_gid(pwd, node);
-	else if (strcmp_1u(node->name, "pgrp") == 0)
+	else if (xml_strcmp(node->name, "pgrp") == 0)
 		return rc_volume_cond_pgrp(pwd, node);
-	else if (strcmp_1u(node->name, "sgrp") == 0)
+	else if (xml_strcmp(node->name, "sgrp") == 0)
 		return rc_volume_cond_sgrp(pwd, node);
 
 	l0g("config: unknown element <%s>\n", node->name);
@@ -1024,12 +1103,12 @@ static int rc_volume_cond_ext(const struct passwd *pwd, xmlNode *node)
  */
 static int rc_volume_cond_simple(const struct passwd *pwd, xmlNode *node)
 {
-	char *user   = xmlGetProp_2s(node, "user");
-	char *invert = xmlGetProp_2s(node, "invert");
-	char *uid    = xmlGetProp_2s(node, "uid");
-	char *gid    = xmlGetProp_2s(node, "gid");
-	char *pgrp   = xmlGetProp_2s(node, "pgrp");
-	char *sgrp   = xmlGetProp_2s(node, "sgrp");
+	char *user   = xml_getprop(node, "user");
+	char *invert = xml_getprop(node, "invert");
+	char *uid    = xml_getprop(node, "uid");
+	char *gid    = xml_getprop(node, "gid");
+	char *pgrp   = xml_getprop(node, "pgrp");
+	char *sgrp   = xml_getprop(node, "sgrp");
 	bool for_me  = true;
 	int ret      = true;
 
@@ -1130,12 +1209,11 @@ static int rc_volume_cond(const char *user, xmlNode *node)
 static const char *rc_volume(xmlNode *node, struct config *config,
     unsigned int command)
 {
+	const char *err;
 	struct vol *vpt;
 	unsigned int i;
 	char *tmp;
 
-	if (strlen(config->user) > sizeof_z(vpt->user))
-		return "config: <volume> components too long\n";
 	if (rc_volume_cond(config->user, node) <= 0)
 		return NULL;
 
@@ -1147,21 +1225,18 @@ static const char *rc_volume(xmlNode *node, struct config *config,
 	HXclist_push(&config->volume_list, &vpt->list);
 
 	vpt->globalconf = config->level == CONTEXT_GLOBAL;
-	strncpy(vpt->user, config->user, sizeof(vpt->user));
+	vpt->user = config->user;
 	vpt->type = CMD_LCLMOUNT;
 	HXclist_init(&vpt->options);
 
 	/* Eyeball ssh setting */
-	if ((tmp = xmlGetProp_2s(node, "ssh")) != NULL) {
-		vpt->uses_ssh = parse_bool(tmp);
-		free(tmp);
-	}
+	if ((tmp = xml_getprop(node, "ssh")) != NULL)
+		vpt->uses_ssh = parse_bool_f(tmp);
 
 	/* Filesystem type */
-	if ((tmp = xmlGetProp_2s(node, "fstype")) != NULL) {
-		if (strlen(tmp) > sizeof_z(vpt->fstype))
-			l0g("config: %s \"%s\" truncated\n", "fstype", tmp);
-		strncpy(vpt->fstype, tmp, sizeof(vpt->fstype));
+	if ((tmp = xml_getprop(node, "fstype")) != NULL) {
+		free(vpt->fstype);
+		vpt->fstype = tmp;
 
 		for (i = 0; default_command[i].type != -1; ++i) {
 			const struct pmt_command *c = &default_command[i];
@@ -1170,127 +1245,130 @@ static const char *rc_volume(xmlNode *node, struct config *config,
 				break;
 			}
 		}
-		free(tmp);
 	} else {
-		strncpy(vpt->fstype, "auto", sizeof(vpt->fstype));
+		vpt->fstype = xstrdup("auto");
+	}
+
+	if ((tmp = xml_getprop(node, "noroot")) != NULL) {
+		vpt->noroot = parse_bool_f(tmp);
+	} else if (vpt->fstype != NULL) {
+		/* Figure out whether we want to act as user. */
+		vpt->noroot =
+			strcmp(vpt->fstype, "fuse") == 0 ||
+			strcmp(vpt->fstype, "encfs13") == 0;
 	}
 
 	/* Source location */
-	if ((tmp = xmlGetProp_2s(node, "server")) != NULL) {
-		if (strlen(tmp) > sizeof_z(vpt->server))
-			l0g("config: %s \"%s\" truncated\n", "server", tmp);
-		strncpy(vpt->server, tmp, sizeof(vpt->server));
-		free(tmp);
+	if ((tmp = xml_getprop(node, "server")) != NULL) {
+		free(vpt->server);
+		vpt->server = tmp;
 	}
-	if ((tmp = xmlGetProp_2s(node, "path")) != NULL) {
-		if (strlen(tmp) > sizeof_z(vpt->volume))
-			l0g("config: %s \"%s\" truncated\n", "path", tmp);
-		strncpy(vpt->volume, tmp, sizeof(vpt->volume));
-		free(tmp);
+	if ((tmp = xml_getprop(node, "path")) != NULL) {
+		free(vpt->volume);
+		vpt->volume = tmp;
 	}
 
 	/* Destination */
-	if ((tmp = xmlGetProp_2s(node, "mountpoint")) != NULL) {
-		if (strlen(tmp) > sizeof_z(vpt->mountpoint))
-			l0g("config: %s \"%s\" truncated\n", "mountpoint", tmp);
-		strncpy(vpt->mountpoint, tmp, sizeof(vpt->mountpoint));
-		free(tmp);
+	if ((tmp = xml_getprop(node, "mountpoint")) != NULL) {
+		free(vpt->mountpoint);
+		vpt->mountpoint = tmp;
 	} else {
-		if (!fstab_value(vpt->volume, FSTAB_MNTPT, vpt->mountpoint,
-		    sizeof(vpt->mountpoint)))
-			return "could not determine mountpoint";
+		free(vpt->mountpoint);
+		vpt->mountpoint = fstab_value(vpt->volume, FSTAB_MNTPT);
+		if (vpt->mountpoint == NULL) {
+			err = "could not determine mountpoint";
+			goto out;
+		}
 		vpt->use_fstab = 1;
 	}
 
 	/* Options */
-	if ((tmp = xmlGetProp_2s(node, "options")) == NULL) {
+	if ((tmp = xml_getprop(node, "options")) == NULL) {
 		/*
 		 * Three options: field defined, field is '-' and fstab should
 		 * be used (when no mount point was provided either) or field
 		 * is '-' and this means no options.
 		 */
 		if (vpt->use_fstab) {
-			char options[MAX_PAR + 1];
-			if (!fstab_value(vpt->volume, FSTAB_OPTS, options,
-			    sizeof(options)))
-				return "could not determine options";
-			if (!str_to_optkv(&vpt->options, options))
-				return "error parsing mount options";
+			char *options = fstab_value(vpt->volume, FSTAB_OPTS);
+			if (options == NULL) {
+				err = "could not determine options";
+				goto out;
+			}
+			if (!str_to_optkv(&vpt->options, options)) {
+				err = "error parsing mount options";
+				goto out;
+			}
+			free(options);
 		}
 	} else if (!str_to_optkv(&vpt->options, tmp)) {
 		free(tmp);
-		return "error parsing mount options";
+		err = "error parsing mount options";
+		goto out;
 	} else {
 		free(tmp);
 	}
 
 	/* Filesystem key */
-	if ((tmp = xmlGetProp_2s(node, "fskeycipher")) != NULL) {
-		if (strlen(tmp) > sizeof_z(vpt->fs_key_cipher))
-			l0g("config: %s \"%s\" truncated\n", "fskeycipher", tmp);
-		strncpy(vpt->fs_key_cipher, tmp, sizeof(vpt->fs_key_cipher));
-		free(tmp);
+	if ((tmp = xml_getprop(node, "cipher")) != NULL) {
+		free(vpt->cipher);
+		vpt->cipher = tmp;
 	}
-	if ((tmp = xmlGetProp_2s(node, "fskeypath")) != NULL) {
-		if (strlen(tmp) > sizeof_z(vpt->fs_key_path))
-			l0g("config: %s \"%s\" truncated\n", "fskeypath", tmp);
-		strncpy(vpt->fs_key_path, tmp, sizeof(vpt->fs_key_path));
-		free(tmp);
+	if ((tmp = xml_getprop(node, "fskeycipher")) != NULL) {
+		free(vpt->fs_key_cipher);
+		vpt->fs_key_cipher = tmp;
+	}
+	if ((tmp = xml_getprop(node, "fskeyhash")) != NULL) {
+		free(vpt->fs_key_hash);
+		vpt->fs_key_hash = tmp;
+	}
+	if ((tmp = xml_getprop(node, "fskeypath")) != NULL) {
+		free(vpt->fs_key_path);
+		vpt->fs_key_path = tmp;
 	}
 
-	/* expandconfig() will set this later */
-	vpt->used_wildcard = false;
 	return NULL;
+
+ out:
+	HXclist_del(&config->volume_list, &vpt->list);
+	volume_free(vpt);
+	return err;
 }
 
 //-----------------------------------------------------------------------------
 static const struct pmt_command default_command[] = {
-	{CMD_SMBMOUNT,   "smbfs", "smbmount",   {"smbmount", "//%(SERVER)/%(VOLUME)", "%(MNTPT)", "-o", "username=%(USER),uid=%(USERUID),gid=%(USERGID)%(before=\",\" OPTIONS)", NULL}},
-	{CMD_SMBUMOUNT,  "smbfs", "smbumount",  {"smbumount", "%(MNTPT)", NULL}},
-	{CMD_CIFSMOUNT,  "cifs",  "cifsmount",  {"mount", "-t", "cifs", "//%(SERVER)/%(VOLUME)", "%(MNTPT)", "-o", "username=%(USER),uid=%(USERUID),gid=%(USERGID)%(before=\",\" OPTIONS)", NULL}},
-	{CMD_NCPMOUNT,   "ncpfs", "ncpmount",   {"ncpmount", "%(SERVER)/%(USER)", "%(MNTPT)", "-o", "pass-fd=0,volume=%(VOLUME)%(before=\",\" OPTIONS)", NULL}},
-	{CMD_NCPUMOUNT,  "ncpfs", "ncpumount",  {"ncpumount", "%(MNTPT)", NULL}},
-	{CMD_FUSEMOUNT,  "fuse",  "fusemount",  {"mount.fuse", "%(VOLUME)", "%(MNTPT)", "%(before=\"-o \" OPTIONS)", NULL}},
-	{CMD_FUSEUMOUNT, "fuse",  "fuseumount", {"fusermount", "-u", "%(MNTPT)", NULL}},
+	{CMD_SMBMOUNT,   "smbfs", {"smbmount", "//%(SERVER)/%(VOLUME)", "%(MNTPT)", "-o", "username=%(USER),uid=%(USERUID),gid=%(USERGID)%(before=\",\" OPTIONS)", NULL}},
+	{CMD_SMBUMOUNT,  "smbfs", {"smbumount", "%(MNTPT)", NULL}},
+	{CMD_CIFSMOUNT,   "cifs", {"mount", "-t", "cifs", "//%(SERVER)/%(VOLUME)", "%(MNTPT)", "-o", "user=%(USER),uid=%(USERUID),gid=%(USERGID)%(before=\",\" OPTIONS)", NULL}},
+	{CMD_NCPMOUNT,   "ncpfs", {"ncpmount", "%(SERVER)/%(USER)", "%(MNTPT)", "-o", "pass-fd=0,volume=%(VOLUME)%(before=\",\" OPTIONS)", NULL}},
+	{CMD_NCPUMOUNT,  "ncpfs", {"ncpumount", "%(MNTPT)", NULL}},
+	{CMD_FUSEMOUNT,   "fuse", {"mount.fuse", "%(VOLUME)", "%(MNTPT)", "%(ifnempty=\"-o\" OPTIONS)", "%(OPTIONS)", NULL}},
+	{CMD_FUSEUMOUNT, "fuse",  {"fusermount", "-u", "%(MNTPT)", NULL}},
 	/* Do not use LCLMOUNT to avoid calling fsck */
-	{CMD_NFSMOUNT,   "nfs",   "nfsmount",   {"mount", "%(SERVER):%(VOLUME)", "%(MNTPT)%(before=\"-o\" OPTIONS)", NULL}},
-	{CMD_LCLMOUNT,   NULL,    "lclmount",   {"mount", "-p0", "-t", "%(FSTYPE)", "%(VOLUME)", "%(MNTPT)", "%(before=\"-o\" OPTIONS)", NULL}},
-	/* Hope to have this in util-linux (LCLMOUNT) some day: */
-	{CMD_CRYPTMOUNT, "crypt", "cryptmount", {"mount.crypt", "%(before=\"-o \" OPTIONS)", "%(VOLUME)", "%(MNTPT)", NULL}},
-	{CMD_CRYPTUMOUNT, "crypt", "cryptumount", {"umount.crypt", "%(MNTPT)", NULL}},
-	{CMD_UMOUNT,     NULL,    "umount",     {"umount", "%(MNTPT)", NULL}},
-	{CMD_LSOF,       NULL,    "lsof",       {"lsof", "%(MNTPT)", NULL}},
-	/*
-	 *  Leave mntcheck available on GNU/Linux so I can ship one
-	 *  config file example
-	 */
-	{CMD_MNTCHECK,   NULL,    "mntcheck",   {"mount", NULL}},
-	{CMD_FSCK,       NULL,    "fsck",       {"fsck", "-p", "%(FSCKTARGET)", NULL}},
-	{CMD_LOSETUP,    NULL,    "losetup",    {"losetup", "-p0", "%(before=\"-e\" CIPHER)", "%(before=\"-k\" KEYBITS)", "%(FSCKLOOP)", "%(VOLUME)", NULL}},
-	{CMD_UNLOSETUP,  NULL,    "unlosetup",  {"losetup", "-d", "%(FSCKLOOP)", NULL}},
-	{CMD_PMVARRUN,   NULL,    "pmvarrun",   {"pmvarrun", "-u", "%(USER)", "-o", "%(OPERATION)", NULL}},
-
-	{CMD_TRUECRYPTMOUNT,  "truecrypt", "truecrypt", {"truecryptmount", "%(VOLUME)", "%(MNTPT)", NULL}},
-	{CMD_TRUECRYPTUMOUNT, "truecrypt", "truecrypt", {"truecryptumount", "-d", "%(MNTPT)", NULL}},
-
+	{CMD_NFSMOUNT,    "nfs",   {"mount", "%(ifnempty=\"-o\" OPTIONS)", "%(OPTIONS)", "-t", "%(FSTYPE)", "%(SERVER):%(VOLUME)", "%(MNTPT)", NULL}},
+	{CMD_NFSMOUNT,    "nfs4"},
+	{CMD_LCLMOUNT,    NULL,    {"mount", "-p0", "%(ifnempty=\"-o\" OPTIONS)", "%(OPTIONS)", "-t", "%(FSTYPE)", "%(VOLUME)", "%(MNTPT)", NULL}},
+	{CMD_CRYPTMOUNT,  "crypt", {"mount.crypt", "%(ifnempty=\"-ocipher=\" CIPHER)%(CIPHER)", "%(ifnempty=\"-ofsk_cipher=\" FSKEYCIPHER)%(FSKEYCIPHER)", "%(ifnempty=\"-ofsk_hash=\" FSKEYHASH)%(FSKEYHASH)", "%(ifnempty=\"-okeyfile=\" FSKEYPATH)%(FSKEYPATH)", "%(ifnempty=\"-o\" OPTIONS)%(OPTIONS)", "%(VOLUME)", "%(MNTPT)", NULL}},
+	{CMD_CRYPTUMOUNT, "crypt", {"umount.crypt", "%(MNTPT)", NULL}},
+	{CMD_UMOUNT,     NULL,     {"umount", "%(MNTPT)", NULL}},
+	{CMD_FSCK,       NULL,     {"fsck", "-p", "%(FSCKTARGET)", NULL}},
+	{CMD_PMVARRUN,   NULL,     {"pmvarrun", "-u", "%(USER)", "-o", "%(OPERATION)", NULL}},
 	{-1},
 };
 
 static const struct callbackmap cf_tags[] = {
 	{"cifsmount",       rc_command,             CMD_CIFSMOUNT},
 	{"cryptmount",      rc_command,             CMD_CRYPTMOUNT},
+	{"cryptumount",     rc_command,             CMD_CRYPTUMOUNT},
 	{"debug",           rc_debug,               CMD_NONE},
 	{"fd0ssh",          rc_command,             CMD_FD0SSH},
-	{"fsckloop",        rc_fsckloop,            CMD_NONE},
 	{"fsck",            rc_command,             CMD_FSCK},
 	{"fusemount",       rc_command,             CMD_FUSEMOUNT},
 	{"fuseumount",      rc_command,             CMD_FUSEUMOUNT},
 	{"lclmount",        rc_command,             CMD_LCLMOUNT},
-	{"losetup",         rc_command,             CMD_LOSETUP},
-	{"lsof",            rc_command,             CMD_LSOF},
+	{"logout",          rc_logout,              CMD_NONE},
 	{"luserconf",       rc_luserconf,           CMD_NONE},
 	{"mkmountpoint",    rc_mkmountpoint,        CMD_NONE},
-	{"mntcheck",        rc_command,             CMD_MNTCHECK},
 	{"mntoptions",      rc_mntoptions,          CMD_NONE},
 	{"msg-authpw",      rc_string,              CMDA_AUTHPW},
 	{"msg-sessionpw",   rc_string,              CMDA_SESSIONPW},
@@ -1301,10 +1379,7 @@ static const struct callbackmap cf_tags[] = {
 	{"pmvarrun",        rc_command,             CMD_PMVARRUN},
 	{"smbmount",        rc_command,             CMD_SMBMOUNT},
 	{"smbumount",       rc_command,             CMD_SMBUMOUNT},
-	{"truecryptmount",  rc_command,             CMD_TRUECRYPTMOUNT},
-	{"truecryptumount", rc_command,             CMD_TRUECRYPTUMOUNT},
 	{"umount",          rc_command,             CMD_UMOUNT},
-	{"unlosetup",       rc_command,             CMD_UNLOSETUP},
 	{"volume",          rc_volume,              CMD_NONE},
 	{NULL},
 };
